import sys

if sys.version_info[:2] >= (3, 8):  # pragma: nocover
    from importlib.metadata import version as importlib_metadata_version
else:  # pragma: nocover
    # Python 3.7 doesn't have this.
    from importlib_metadata import version as importlib_metadata_version


__version__ = importlib_metadata_version(__package__)



"""

SporeStack API supplemental models

"""


from typing import Optional

from pydantic import BaseModel


class NetworkInterface(BaseModel):
    ipv4: str
    ipv6: str


class Payment(BaseModel):
    txid: Optional[str]
    uri: Optional[str]
    usd: str
    paid: bool



import secrets
from base64 import b16encode
from struct import pack
from zlib import adler32


def checksum(to_hash: str) -> str:
    """
    Base 16 string of half the adler32 checksum
    """
    adler32_hash = adler32(bytes(to_hash, "utf-8"))
    return b16encode(pack("I", adler32_hash)).decode("utf-8").lower()[-4:]


def random_machine_id() -> str:
    """
    These used to be 64 hex characters. Now they have a new format.
    """
    to_hash = f"ss_m_{secrets.token_hex(11)}"
    return f"{to_hash}_{checksum(to_hash)}"


def random_token() -> str:
    """
    64 hex characters.
    """
    return secrets.token_hex(32)



"""
SporeStack CLI: `sporestack`
"""

import importlib.util
import json
import logging
import os
import sys
import time
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Dict, Optional

if sys.version_info[:2] >= (3, 8):  # pragma: nocover
    from importlib.metadata import version as importlib_metadata_version
else:  # pragma: nocover
    # Python 3.7 doesn't have this.
    from importlib_metadata import version as importlib_metadata_version

import typer


def lazy_import(name: str) -> ModuleType:
    """
    Lazily import a module. Helps speed up CLI performance.
    """
    spec = importlib.util.find_spec(name)
    assert spec is not None
    assert spec.loader is not None
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


# For mypy
if TYPE_CHECKING:
    from . import api_client
else:
    api_client = lazy_import("sporestack.api_client")

HELP = """
SporeStack Python CLI

Optional environment variables:
SPORESTACK_ENDPOINT
*or*
SPORESTACK_USE_TOR_ENDPOINT

TOR_PROXY (defaults to socks5h://127.0.0.1:9050 which is fine for most)
"""

_home = os.getenv("HOME", None)
assert _home is not None, "Unable to detect $HOME environment variable?"
HOME = Path(_home)

SPORESTACK_DIR = HOME / ".sporestack"

# Try to protect files in ~/.sporestack
os.umask(0o0077)

cli = typer.Typer(help=HELP)

token_cli = typer.Typer()
HOME = Path(_home)
cli.add_typer(token_cli, name="token")
server_cli = typer.Typer()
cli.add_typer(server_cli, name="server")

logging.basicConfig(level=logging.INFO)

DEFAULT_TOKEN = "primary"
DEFAULT_FLAVOR = "vps-1vcpu-1gb"
# Users may have a different key file, but this is the most common.
DEFAULT_SSH_KEY_FILE = HOME / ".ssh" / "id_rsa.pub"

# On disk format
TOKEN_VERSION = 1

WAITING_PAYMENT_TO_PROCESS = "Waiting for payment to process..."


def get_api_endpoint() -> str:
    api_endpoint = os.getenv("SPORESTACK_ENDPOINT", api_client.CLEARNET_ENDPOINT)
    if os.getenv("SPORESTACK_USE_TOR_ENDPOINT", None) is not None:
        api_endpoint = api_client.TOR_ENDPOINT
    return api_endpoint


def make_payment(currency: str, uri: str, usd: str) -> None:
    import segno

    premessage = """Payment URI: {}
Pay *exactly* the specified amount. No more, no less. Pay within
one hour at the very most.
Resize your terminal and try again if QR code above is not readable.
Press ctrl+c to abort."""
    message = premessage.format(uri)
    qr = segno.make(uri)
    # This typer.echos.
    qr.terminal()
    typer.echo(message)
    typer.echo(f"Approximate price in USD: {usd}")
    input("[Press enter once you have made payment.]")


@server_cli.command()
def launch(
    hostname: str,
    days: int = typer.Option(...),
    operating_system: str = typer.Option(...),
    ssh_key_file: Path = DEFAULT_SSH_KEY_FILE,
    flavor: str = DEFAULT_FLAVOR,
    token: str = DEFAULT_TOKEN,
    region: Optional[str] = None,
    quote: bool = typer.Option(True, help="Require manual price confirmation."),
) -> None:
    """
    Launch a server on SporeStack.
    """

    from . import utils

    typer.echo(f"Launching server with token {token}...", err=True)
    _token = load_token(token)

    if machine_exists(hostname):
        typer.echo(f"{hostname} already created.")
        raise typer.Exit(code=1)

    typer.echo(f"Loading SSH key from {ssh_key_file}...")
    if not ssh_key_file.exists():
        msg = f"{ssh_key_file} does not exist. "
        msg += "You can try generating a key file with `ssh-keygen`"
        typer.echo(msg, err=True)
        raise typer.Exit(code=1)

    ssh_key = ssh_key_file.read_text()

    machine_id = utils.random_machine_id()

    if quote:
        response = api_client.launch(
            machine_id=machine_id,
            days=days,
            flavor=flavor,
            operating_system=operating_system,
            ssh_key=ssh_key,
            currency="settlement",
            region=region,
            token=_token,
            api_endpoint=get_api_endpoint(),
            retry=True,
            quote=True,
        )

        msg = f"Is {response.payment.usd} for {days} day(s) of {flavor} okay?"
        typer.echo(msg, err=True)
        input("[Press ctrl+c to cancel, or enter to accept.]")

    tries = 360
    while tries > 0:
        response = api_client.launch(
            machine_id=machine_id,
            days=days,
            flavor=flavor,
            operating_system=operating_system,
            ssh_key=ssh_key,
            currency="settlement",
            region=region,
            token=_token,
            api_endpoint=get_api_endpoint(),
            retry=True,
        )
        if response.created is True:
            break
        typer.echo("Waiting for server to build...", err=True)
        tries = tries + 1
        # Waiting for server to spin up.
        time.sleep(10)

    if response.created is False:
        typer.echo("Server creation failed, tries exceeded.", err=True)
        raise typer.Exit(code=1)

    created_dict = response.dict()
    created_dict["vm_hostname"] = hostname
    save_machine_info(created_dict)
    typer.echo(pretty_machine_info(created_dict), err=True)
    typer.echo(json.dumps(created_dict, indent=4))


@server_cli.command()
def topup(
    hostname: str,
    days: int = typer.Option(...),
    token: str = DEFAULT_TOKEN,
    quote: bool = typer.Option(True, help="Require manual price confirmation."),
) -> None:
    """
    Extend an existing SporeStack server's lifetime.
    """

    if not machine_exists(hostname):
        typer.echo(f"{hostname} does not exist.")
        raise typer.Exit(code=1)

    _token = load_token(token)

    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]

    if quote:
        response = api_client.topup(
            machine_id=machine_id,
            days=days,
            currency="settlement",
            api_endpoint=get_api_endpoint(),
            token=_token,
            retry=True,
            quote=True,
        )

        typer.echo(f"Is {response.payment.usd} for {days} day(s) okay?", err=True)
        input("[Press ctrl+c to cancel, or enter to accept.]")

    response = api_client.topup(
        machine_id=machine_id,
        days=days,
        currency="settlement",
        api_endpoint=get_api_endpoint(),
        token=_token,
        retry=True,
    )
    assert response.payment.paid is True

    machine_info["expiration"] = response.expiration
    save_machine_info(machine_info, overwrite=True)
    typer.echo(machine_info["expiration"])


def server_info_path() -> Path:
    # Put servers in a subdirectory
    servers_dir = SPORESTACK_DIR / "servers"

    # Migrate existing server.json files into servers subdirectory
    if SPORESTACK_DIR.exists() and not servers_dir.exists():
        typer.echo(
            f"Migrating server profiles found in {SPORESTACK_DIR} to {servers_dir}.",
            err=True,
        )
        servers_dir.mkdir()
        for json_file in SPORESTACK_DIR.glob("*.json"):
            json_file.rename(servers_dir / json_file.name)

    # Make it, if it doesn't exist already.
    SPORESTACK_DIR.mkdir(exist_ok=True)
    servers_dir.mkdir(exist_ok=True)

    return servers_dir


def token_path() -> Path:
    token_dir = SPORESTACK_DIR / "tokens"

    # Make it, if it doesn't exist already.
    token_dir.mkdir(exist_ok=True, parents=True)

    return token_dir


def save_machine_info(machine_info: Dict[str, Any], overwrite: bool = False) -> None:
    """
    Save info to disk.
    """
    directory = server_info_path()
    hostname = machine_info["vm_hostname"]
    json_file = directory / f"{hostname}.json"
    if overwrite is False:
        assert json_file.exists() is False, f"{json_file} already exists."
    json_file.write_text(json.dumps(machine_info))


def get_machine_info(hostname: str) -> Dict[str, Any]:
    """
    Get info from disk.
    """
    directory = server_info_path()
    json_file = directory / f"{hostname}.json"
    if not json_file.exists():
        raise ValueError(f"{hostname} does not exist in {directory} as {json_file}")
    machine_info = json.loads(json_file.read_bytes())
    assert isinstance(machine_info, dict)
    if machine_info["vm_hostname"] != hostname:
        raise ValueError("hostname does not match filename.")
    return machine_info


def pretty_machine_info(info: Dict[str, Any]) -> str:
    msg = "Hostname: {}\n".format(info["vm_hostname"])
    msg += "Machine ID (keep this secret!): {}\n".format(info["machine_id"])
    if "ipv6" in info["network_interfaces"][0]:
        msg += "IPv6: {}\n".format(info["network_interfaces"][0]["ipv6"])
    if "ipv4" in info["network_interfaces"][0]:
        msg += "IPv4: {}\n".format(info["network_interfaces"][0]["ipv4"])
    expiration = info["expiration"]
    human_expiration = time.strftime("%Y-%m-%d %H:%M:%S %z", time.localtime(expiration))
    if "running" in info:
        msg += "Running: {}\n".format(info["running"])
    msg += f"Expiration: {expiration} ({human_expiration})\n"
    time_to_live = expiration - int(time.time())
    hours = time_to_live // 3600
    msg += f"Server will be deleted in {hours} hours."
    return msg


@server_cli.command(name="list")
def server_list() -> None:
    """
    List all locally known servers.
    """
    from .exceptions import SporeStackUserError

    directory = server_info_path()
    infos = []
    for hostname_json in os.listdir(directory):
        hostname = hostname_json.split(".")[0]
        saved_vm_info = get_machine_info(hostname)
        try:
            upstream_vm_info = api_client.info(
                machine_id=saved_vm_info["machine_id"],
                api_endpoint=get_api_endpoint(),
            )
            saved_vm_info["expiration"] = upstream_vm_info.expiration
            saved_vm_info["running"] = upstream_vm_info.running
            infos.append(saved_vm_info)
        except SporeStackUserError as e:
            expiration = saved_vm_info["expiration"]
            human_expiration = time.strftime(
                "%Y-%m-%d %H:%M:%S %z", time.localtime(expiration)
            )
            msg = hostname
            msg += f" expired ({expiration} {human_expiration}): "
            msg += str(e)
            typer.echo(msg)

    for info in infos:
        typer.echo()
        typer.echo(pretty_machine_info(info))

    typer.echo()


def machine_exists(hostname: str) -> bool:
    """
    Check if the VM's JSON exists locally.
    """
    return server_info_path().joinpath(f"{hostname}.json").exists()


@server_cli.command()
def get_attribute(hostname: str, attribute: str) -> None:
    """
    Returns an attribute about the VM.
    """
    machine_info = get_machine_info(hostname)
    typer.echo(machine_info[attribute])


@server_cli.command()
def info(hostname: str) -> None:
    """
    Info on the VM
    """
    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]
    typer.echo(
        api_client.info(machine_id=machine_id, api_endpoint=get_api_endpoint()).json()
    )


@server_cli.command()
def start(hostname: str) -> None:
    """
    Boots the VM.
    """
    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]
    api_client.start(machine_id=machine_id, api_endpoint=get_api_endpoint())
    typer.echo(f"{hostname} started.")


@server_cli.command()
def stop(hostname: str) -> None:
    """
    Immediately kills the VM.
    """
    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]
    api_client.stop(machine_id=machine_id, api_endpoint=get_api_endpoint())
    typer.echo(f"{hostname} stopped.")


@server_cli.command()
def delete(hostname: str) -> None:
    """
    Deletes the VM before expiration (no refunds/credits)
    """
    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]
    api_client.delete(machine_id=machine_id, api_endpoint=get_api_endpoint())
    # Also remove the .json file
    server_info_path().joinpath(f"{hostname}.json").unlink()
    typer.echo(f"{hostname} was deleted.")


@server_cli.command()
def rebuild(hostname: str) -> None:
    """
    Rebuilds the VM with the operating system and SSH key given at launch time.

    Will take a couple minutes to complete after the request is made.
    """
    machine_info = get_machine_info(hostname)
    machine_id = machine_info["machine_id"]
    api_client.rebuild(machine_id=machine_id, api_endpoint=get_api_endpoint())
    typer.echo(f"{hostname} rebuilding.")


def load_token(token: str) -> str:
    token_file = token_path().joinpath(f"{token}.json")
    if not token_file.exists():
        msg = f"Token '{token}' ({token_file}) does not exist. Create it with:\n"
        msg += f"sporestack token create {token} --dollars 20 --currency xmr\n"
        msg += "(Can do more than $20, or a different currency, like btc.)\n"
        msg += (
            "With the token credited, you can launch servers, renew existing ones, etc."
        )
        typer.echo(msg, err=True)
        raise typer.Exit(code=1)

    token_data = json.loads(token_file.read_text())
    assert token_data["version"] == 1
    assert isinstance(token_data["key"], str)
    return token_data["key"]


def save_token(token: str, key: str) -> None:
    token_file = token_path().joinpath(f"{token}.json")
    if token_file.exists():
        msg = f"Token '{token}' already exists in {token_file}. Aborting!"
        typer.echo(msg, err=True)
        raise typer.Exit(code=1)

    token_data = {"version": TOKEN_VERSION, "name": token, "key": key}
    token_file.write_text(json.dumps(token_data))


@token_cli.command(name="create")
def token_create(
    token: str = typer.Argument(DEFAULT_TOKEN),
    dollars: int = typer.Option(...),
    currency: str = typer.Option(...),
) -> None:
    """
    Enables a new settlement token.

    Dollars is starting balance.
    """
    from . import utils

    _token = utils.random_token()

    typer.echo(f"Generated key {_token} for use with token {token}", err=True)

    if Path(SPORESTACK_DIR / "tokens" / f"{token}.json").exists():
        typer.echo("Token already created! Did you mean to `topup`?", err=True)
        raise typer.Exit(1)

    response = api_client.token_add(
        token=_token,
        dollars=dollars,
        currency=currency,
        api_endpoint=get_api_endpoint(),
        retry=True,
    )

    uri = response.payment.uri
    assert uri is not None
    usd = response.payment.usd

    make_payment(currency=currency, uri=uri, usd=usd)

    tries = 360 * 2
    while tries > 0:
        typer.echo(WAITING_PAYMENT_TO_PROCESS, err=True)
        tries = tries - 1
        # FIXME: Wait two hours in a smarter way.
        # Waiting for payment to set in.
        time.sleep(10)
        response = api_client.token_add(
            token=_token,
            dollars=dollars,
            currency=currency,
            api_endpoint=get_api_endpoint(),
            retry=True,
        )
        if response.payment.paid is True:
            typer.echo(f"{token} has been enabled with ${dollars}.")
            typer.echo(f"{token}'s key is {_token}.")
            typer.echo("Save it, don't share it, and don't lose it!")
            save_token(token, _token)
            return
    raise ValueError(f"{token} did not get enabled in time.")


@token_cli.command(name="import")
def token_import(
    name: str = typer.Argument(DEFAULT_TOKEN),
    key: str = typer.Option(...),
) -> None:
    """
    Imports a token from a key.
    """
    save_token(name, key)


@token_cli.command(name="topup")
def token_topup(
    token: str = typer.Argument(DEFAULT_TOKEN),
    dollars: int = typer.Option(...),
    currency: str = typer.Option(...),
) -> None:
    """
    Adds balance to an existing settlement token.
    """
    token = load_token(token)

    response = api_client.token_add(
        token,
        dollars,
        currency=currency,
        api_endpoint=get_api_endpoint(),
        retry=True,
    )

    uri = response.payment.uri
    assert uri is not None
    usd = response.payment.usd

    make_payment(currency=currency, uri=uri, usd=usd)

    tries = 360 * 2
    while tries > 0:
        typer.echo(WAITING_PAYMENT_TO_PROCESS, err=True)
        tries = tries - 1
        # FIXME: Wait two hours in a smarter way.
        response = api_client.token_add(
            token,
            dollars,
            currency=currency,
            api_endpoint=get_api_endpoint(),
            retry=True,
        )
        # Waiting for payment to set in.
        time.sleep(10)
        if response.payment.paid is True:
            typer.echo(f"Added {dollars} dollars to {token}")
            return
    raise ValueError(f"{token} did not get enabled in time.")


@token_cli.command()
def balance(token: str = typer.Argument(DEFAULT_TOKEN)) -> None:
    """
    Gets balance for a settlement token.
    """
    _token = load_token(token)

    typer.echo(
        api_client.token_balance(token=_token, api_endpoint=get_api_endpoint()).usd
    )


@token_cli.command(name="list")
def token_list() -> None:
    """
    Gets balance for a settlement token.
    """
    token_dir = token_path()
    typer.echo(f"SporeStack tokens present in {token_dir}:", err=True)
    typer.echo("(Name): (Key)", err=True)
    for token_file in token_dir.glob("*.json"):
        token = token_file.stem
        key = load_token(token)
        typer.echo(f"{token}: {key}")


@cli.command()
def version() -> None:
    """
    Returns the installed version.
    """
    typer.echo(importlib_metadata_version(__package__))


@cli.command()
def api_endpoint() -> None:
    """
    Prints the selected API endpoint: Env var: SPORESTACK_ENDPOINT,
    or, SPORESTACK_USE_TOR=1
    """
    endpoint = get_api_endpoint()
    if ".onion" in endpoint:
        typer.echo(f"{endpoint} using {api_client._get_tor_proxy()}")
        return
    else:
        typer.echo(endpoint)
        return


if __name__ == "__main__":
    cli()



import logging
import os
from time import sleep
from typing import Any, Dict, Optional

import requests

from . import api, exceptions
from .version import __version__

log = logging.getLogger(__name__)

LATEST_API_VERSION = 2

CLEARNET_ENDPOINT = "https://api.sporestack.com"
TOR_ENDPOINT = (
    "http://api.spore64i5sofqlfz5gq2ju4msgzojjwifls7rok2cti624zyq3fcelad.onion"
)

API_ENDPOINT = CLEARNET_ENDPOINT

GET_TIMEOUT = 60
POST_TIMEOUT = 90
USE_TOR_PROXY = "auto"


def _get_tor_proxy() -> str:
    """
    This makes testing easier.
    """
    return os.getenv("TOR_PROXY", "socks5h://127.0.0.1:9050")


# For requests module
TOR_PROXY_REQUESTS = {"http": _get_tor_proxy(), "https": _get_tor_proxy()}


def _is_onion_url(url: str) -> bool:
    """
    returns True/False depending on if a URL looks like a Tor hidden service
    (.onion) or not.
    This is designed to false as non-onion just to be on the safe-ish side,
    depending on your view point. It requires URLs like: http://domain.tld/,
    not http://domain.tld or domain.tld/.

    This can be optimized a lot.
    """
    try:
        url_parts = url.split("/")
        domain = url_parts[2]
        tld = domain.split(".")[-1]
        if tld == "onion":
            return True
    except Exception:
        pass
    return False


def _api_request(
    url: str,
    empty_post: bool = False,
    json_params: Optional[Dict[str, Any]] = None,
    retry: bool = False,
) -> Any:
    headers = {"User-Agent": f"sporestack-python/{__version__}"}
    proxies = {}
    if _is_onion_url(url) is True:
        log.debug("Got a .onion API endpoint, using local Tor SOCKS proxy.")
        proxies = TOR_PROXY_REQUESTS

    try:
        if empty_post is True:
            request = requests.post(
                url, timeout=POST_TIMEOUT, proxies=proxies, headers=headers
            )
        elif json_params is None:
            request = requests.get(
                url, timeout=GET_TIMEOUT, proxies=proxies, headers=headers
            )
        else:
            request = requests.post(
                url,
                json=json_params,
                timeout=POST_TIMEOUT,
                proxies=proxies,
                headers=headers,
            )
    except Exception as e:
        if retry is True:
            log.warning(f"Got an error, but retrying: {e}")
            sleep(5)
            # Try again.
            return _api_request(
                url,
                empty_post=empty_post,
                json_params=json_params,
                retry=retry,
            )
        else:
            raise

    status_code_first_digit = request.status_code // 100
    if status_code_first_digit == 2:
        try:
            return request.json()
        except Exception:
            return request.content
    elif status_code_first_digit == 4:
        log.debug("HTTP status code: {request.status_code}")
        raise exceptions.SporeStackUserError(request.content.decode("utf-8"))
    elif status_code_first_digit == 5:
        if retry is True:
            log.warning(request.content.decode("utf-8"))
            log.warning("Got a 500, retrying in 5 seconds...")
            sleep(5)
            # Try again if we get a 500
            return _api_request(
                url,
                empty_post=empty_post,
                json_params=json_params,
                retry=retry,
            )
        else:
            raise exceptions.SporeStackServerError(str(request.content))
    else:
        # Not sure why we'd get this.
        request.raise_for_status()
        raise Exception("Stuff broke strangely. Please contact SporeStack support.")


def launch(
    machine_id: str,
    days: int,
    currency: str,
    flavor: str,
    operating_system: str,
    ssh_key: str,
    api_endpoint: str = API_ENDPOINT,
    region: Optional[str] = None,
    token: Optional[str] = None,
    retry: bool = False,
    affiliate_token: Optional[str] = None,
    quote: bool = False,
) -> api.ServerLaunch.Response:
    request = api.ServerLaunch.Request(
        machine_id=machine_id,
        days=days,
        currency=currency,
        token=token,
        affiliate_token=affiliate_token,
        flavor=flavor,
        region=region,
        operating_system=operating_system,
        ssh_key=ssh_key,
        quote=quote,
    )
    url = api_endpoint + api.ServerLaunch.url.format(machine_id=machine_id)
    response = _api_request(url=url, json_params=request.dict(), retry=retry)
    response_object = api.ServerLaunch.Response.parse_obj(response)
    assert response_object.machine_id == machine_id
    return response_object


def topup(
    machine_id: str,
    days: int,
    currency: str,
    api_endpoint: str = API_ENDPOINT,
    token: Optional[str] = None,
    retry: bool = False,
    affiliate_token: Optional[str] = None,
    quote: bool = False,
) -> api.ServerTopup.Response:
    """
    Topup a server.
    """
    request = api.ServerTopup.Request(
        machine_id=machine_id,
        days=days,
        currency=currency,
        token=token,
        affiliate_token=affiliate_token,
        quote=quote,
    )
    url = api_endpoint + api.ServerTopup.url.format(machine_id=machine_id)
    response = _api_request(url=url, json_params=request.dict(), retry=retry)
    response_object = api.ServerTopup.Response.parse_obj(response)
    assert response_object.machine_id == machine_id
    return response_object


def start(machine_id: str, api_endpoint: str = API_ENDPOINT) -> None:
    """
    Boots the server.
    """
    url = api_endpoint + api.ServerStart.url.format(machine_id=machine_id)
    _api_request(url, empty_post=True)


def stop(machine_id: str, api_endpoint: str = API_ENDPOINT) -> None:
    """
    Powers off the server.
    """
    url = api_endpoint + api.ServerStop.url.format(machine_id=machine_id)
    _api_request(url, empty_post=True)


def delete(machine_id: str, api_endpoint: str = API_ENDPOINT) -> None:
    """
    Deletes the server.
    """
    url = api_endpoint + api.ServerDelete.url.format(machine_id=machine_id)
    _api_request(url, empty_post=True)


def rebuild(machine_id: str, api_endpoint: str = API_ENDPOINT) -> None:
    """
    Rebuilds the server with the operating system and SSH key set at launch time.

    Deletes all of the data on the server!
    """
    url = api_endpoint + api.ServerRebuild.url.format(machine_id=machine_id)
    _api_request(url, empty_post=True)


def info(machine_id: str, api_endpoint: str = API_ENDPOINT) -> api.ServerInfo.Response:
    """
    Returns info about the server.
    """
    url = api_endpoint + api.ServerInfo.url.format(machine_id=machine_id)
    response = _api_request(url)
    response_object = api.ServerInfo.Response.parse_obj(response)
    assert response_object.machine_id == machine_id
    return response_object


def token_add(
    token: str,
    dollars: int,
    currency: str,
    api_endpoint: str = API_ENDPOINT,
    retry: bool = False,
) -> api.TokenAdd.Response:
    request = api.TokenAdd.Request(dollars=dollars, currency=currency)
    url = api_endpoint + api.TokenAdd.url.format(token=token)
    response = _api_request(url=url, json_params=request.dict(), retry=retry)
    response_object = api.TokenAdd.Response.parse_obj(response)
    assert response_object.token == token
    return response_object


def token_balance(
    token: str, api_endpoint: str = API_ENDPOINT
) -> api.TokenBalance.Response:
    url = api_endpoint + api.TokenBalance.url.format(token=token)
    response = _api_request(url=url)
    response_object = api.TokenBalance.Response.parse_obj(response)
    assert response_object.token == token
    return response_object



"""

SporeStack API request/response models

"""


from typing import List, Optional

from pydantic import BaseModel

from .models import NetworkInterface, Payment

LATEST_API_VERSION = 3


class TokenEnable:
    """Deprecated: Use TokenAdd instead."""

    url = "/token/{token}/enable"
    method = "POST"

    class Request(BaseModel):
        currency: str
        dollars: int
        affiliate_token: Optional[str] = None

    class Response(BaseModel):
        token: str
        payment: Payment


class TokenAdd:
    url = "/token/{token}/add"
    method = "POST"

    class Request(BaseModel):
        currency: str
        dollars: int
        affiliate_token: Optional[str] = None

    class Response(BaseModel):
        token: str
        payment: Payment


class TokenBalance:
    url = "/token/{token}/balance"
    method = "GET"

    class Response(BaseModel):
        token: str
        cents: int
        usd: str


class ServerLaunch:
    url = "/server/{machine_id}/launch"
    method = "POST"

    class Request(BaseModel):
        machine_id: str
        days: int
        flavor: str
        ssh_key: str
        operating_system: str
        currency: Optional[str] = None
        """Currency only needs to be set if not paying with a token."""
        region: Optional[str] = None
        organization: Optional[str] = None
        token: Optional[str] = None
        quote: bool = False
        affiliate_token: Optional[str] = None
        affiliate_amount: None = None
        """Deprecated field"""
        settlement_token: Optional[str] = None
        """Deprecated field. Use token instead."""

    class Response(BaseModel):
        payment: Payment
        expiration: int
        machine_id: str
        operating_system: str
        flavor: str
        network_interfaces: List[NetworkInterface] = []
        created_at: int = 0
        region: Optional[str] = None
        latest_api_version: int = LATEST_API_VERSION
        created: bool = False
        paid: bool = False
        warning: Optional[str] = None
        txid: Optional[str] = None


class ServerTopup:
    url = "/server/{machine_id}/topup"
    method = "POST"

    class Request(BaseModel):
        machine_id: str
        days: int
        token: Optional[str] = None
        quote: bool = False
        currency: Optional[str] = None
        """Currency only needs to be set if not paying with a token."""
        affiliate_token: Optional[str] = None
        affiliate_amount: None = None
        """Deprecated field"""
        settlement_token: Optional[str] = None
        """Deprecated field. Use token instead."""

    class Response(BaseModel):
        machine_id: str
        payment: Payment
        expiration: int
        paid: bool = False
        warning: Optional[str] = None
        txid: Optional[str] = None
        latest_api_version: int = LATEST_API_VERSION


class ServerInfo:
    url = "/server/{machine_id}/info"
    method = "GET"

    class Response(BaseModel):
        created_at: int
        expiration: int
        running: bool
        machine_id: str
        network_interfaces: List[NetworkInterface]
        region: str


class ServerStart:
    url = "/server/{machine_id}/start"
    method = "POST"


class ServerStop:
    url = "/server/{machine_id}/stop"
    method = "POST"


class ServerDelete:
    url = "/server/{machine_id}/delete"
    method = "POST"


class ServerRebuild:
    url = "/server/{machine_id}/rebuild"
    method = "POST"
